use crate::drivers::vga::{Vga13h, VgaColor};

use super::FallingTetromino;

/// The state of a cell in the game board
#[derive(Debug, Clone, Copy, PartialEq, Eq)]
pub enum TetrisCell {
	Empty,
	Filled { color: VgaColor },
}

/// The tetris board
pub struct TetrisBoard {
	board: [TetrisCell; TetrisBoard::BOARD_WIDTH * TetrisBoard::BOARD_HEIGHT],
}

impl TetrisBoard {
	/// The width of this board, in cells
	const BOARD_WIDTH: usize = 10;

	/// The height of this board, in cells
	const BOARD_HEIGHT: usize = 20;

	/// The side length of a (square) cell, in pixels
	const CELL_SIZE: usize = 9;

	pub const fn new() -> Self {
		Self {
			board: [TetrisCell::Empty; TetrisBoard::BOARD_WIDTH * TetrisBoard::BOARD_HEIGHT],
		}
	}

	/// Find and remove all filled rows,
	/// shifting upper rows down.
	pub fn collapse(&mut self) {
		let mut y = Self::BOARD_HEIGHT - 1;
		'outer: loop {
			for x in 0..Self::BOARD_WIDTH {
				let cell = self.get_cell(x, y);
				if cell == Some(&TetrisCell::Empty) {
					if y == 0 {
						break 'outer;
					}

					y -= 1;
					continue 'outer;
				}
			}

			// We found a row that needs to be cleared

			// Shift everything down
			for yy in (1..=y).rev() {
				for x in 0..Self::BOARD_WIDTH {
					let top = *self.get_cell(x, yy - 1).unwrap();
					let bot = self.get_cell_mut(x, yy).unwrap();
					*bot = top;
				}
			}

			// Clear the top row
			for x in 0..Self::BOARD_WIDTH {
				*self.get_cell_mut(x, 0).unwrap() = TetrisCell::Empty;
			}
		}
	}

	/// Place the given tetromino on the board,
	/// filling the cells it occupies.
	///
	/// If the tetromino cells that overlap
	/// non-empty board cells are ignored.
	pub fn place_tetromino(&mut self, tetromino: FallingTetromino) {
		for (x, y) in tetromino.tiles() {
			let cell = self.get_cell_mut(x, y);
			if let Some(cell) = cell {
				*cell = TetrisCell::Filled {
					color: tetromino.color,
				};
			}
		}
	}

	/// Returns `false` if the given tetromino intersects a filled cell
	/// or exits the board boundary
	pub fn tetromino_valid(&self, tetromino: &FallingTetromino) -> bool {
		for (x, y) in tetromino.tiles() {
			let cell = self.get_cell(x, y);
			if cell != Some(&TetrisCell::Empty) {
				return false;
			}
		}

		return true;
	}

	/// Get the value of the cell at the given position.
	/// Returns [`None`] if (x, y) exceeds the board's bounds.
	pub fn get_cell(&self, x: usize, y: usize) -> Option<&TetrisCell> {
		return self.board.get(y * TetrisBoard::BOARD_WIDTH + x);
	}

	/// Get a mutable reference to the cell at the given position.
	/// Returns [`None`] if (x, y) exceeds the board's bounds.
	pub fn get_cell_mut(&mut self, x: usize, y: usize) -> Option<&mut TetrisCell> {
		return self.board.get_mut(y * TetrisBoard::BOARD_WIDTH + x);
	}
}

//
// MARK: draw routines
//

impl TetrisBoard {
	/// Draw a cell of the given color on `fb`.
	/// (x, y) is the pixel position of the cell (NOT board coordinates).
	fn draw_cell(&self, fb: &mut [u8], color: VgaColor, x: usize, y: usize) {
		let color = color.as_u8();
		for yo in 0..TetrisBoard::CELL_SIZE {
			let left = Vga13h::pix_idx(x, y + yo);
			let right = Vga13h::pix_idx(x + TetrisBoard::CELL_SIZE, y + yo);
			fb[left..right].copy_from_slice(&[color; TetrisBoard::CELL_SIZE]);
		}
	}

	/// Draw the tetris board's frame
	fn draw_frame(&self, fb: &mut [u8], x: usize, y: usize) {
		let color = VgaColor::Gray.as_u8();
		for yo in 0..TetrisBoard::CELL_SIZE {
			let left = Vga13h::pix_idx(x, y + yo);
			let right = Vga13h::pix_idx(x + TetrisBoard::CELL_SIZE, y + yo);
			fb[left..right].copy_from_slice(&[color; TetrisBoard::CELL_SIZE]);
		}
	}

	/// Draw this tetris board using the given VGA driver.
	pub fn draw(&self, vga: &mut Vga13h, falling: Option<&FallingTetromino>) {
		let fb = vga.get_fb();

		// Draw cells
		for bx in 0..TetrisBoard::BOARD_WIDTH {
			for by in 0..TetrisBoard::BOARD_HEIGHT {
				let cell = self.board[by * TetrisBoard::BOARD_WIDTH + bx];
				let dx = (bx + 1) * TetrisBoard::CELL_SIZE;
				let dy = (by + 1) * TetrisBoard::CELL_SIZE;

				if let TetrisCell::Filled { color } = cell {
					self.draw_cell(fb, color, dx, dy);
				} else {
					self.draw_cell(fb, VgaColor::Black, dx, dy);
				}
			}
		}

		// Draw falling tetromino
		if let Some(falling) = falling {
			for (x, y) in falling.tiles() {
				let dx = (x + 1) * TetrisBoard::CELL_SIZE;
				let dy = (y + 1) * TetrisBoard::CELL_SIZE;

				self.draw_cell(fb, falling.color, dx, dy);
			}
		}

		// Draw frame
		for bx in 0..TetrisBoard::BOARD_WIDTH + 2 {
			self.draw_frame(fb, bx * TetrisBoard::CELL_SIZE, 0);
			self.draw_frame(
				fb,
				bx * TetrisBoard::CELL_SIZE,
				(TetrisBoard::BOARD_HEIGHT + 1) * TetrisBoard::CELL_SIZE,
			);
		}
		for by in 0..TetrisBoard::BOARD_HEIGHT + 2 {
			self.draw_frame(fb, 0, by * 9);
			self.draw_frame(
				fb,
				(TetrisBoard::BOARD_WIDTH + 1) * TetrisBoard::CELL_SIZE,
				by * TetrisBoard::CELL_SIZE,
			);
		}
	}
}
